// SPDX-FileCopyrightText: Deutsches Elektronen-Synchrotron DESY, MSK
// SPDX-License-Identifier: LGPL-3.0-or-later

#include "TecModule.h"

#include "TecErrorCode.h"
#include "TecParameters.h"

#include <regex>

namespace detail {
  static std::string commandToDescription(const std::string& command) {
    if(command == "write" || command == "read") return command + " of ";
    if(command == "setRb") return "set from readback for ";
    if(command == "writeToFlash") return "write to FLASH of ";
    assert(false);
  }

  std::string generateGroupDescription(const std::string& groupName) {
    std::regex re{"(read|write|writeToFlash|setRb)(Ch([0-9]+)|Autotune)?((Write)?(Autotune)?Group)([0-9]*)"};
    std::smatch match;
    if(std::regex_match(groupName, match, re)) {
      assert(match.size() == 8);

      std::string description = "Trigger " + commandToDescription(match[1]);
      if(match[2].matched) {
        if(match[3].matched)
          description += ("channel " + match[3].str() + " ");
        else
          description += "autotune ";
      }
      else
        description += "generic ";

      if(match[1] == "read") {
        description += "read-only parameters group";
      } else
        description += "read-write group";

      if (match[7].matched)
        description += (" " + match[7].str());
    }
    else {
      assert(false);
    }

    return "description";
  }
} // namespace detail

TecModule::TecModule(EntityOwner* owner, const std::string& name, const std::string& description,
    ctk::HierarchyModifier hierarchyModifier, const std::unordered_set<std::string>& tags)
: ctk::ApplicationModule(owner, name, description, hierarchyModifier, tags) {
  assert(checkDescriptorsList());
  createReadWriteMaps();
  createCommandGroupMaps();
  createEnablePeriodicList();
}

/**********************************************************************************************************************/

void TecModule::mainLoop() {
  // enableDebug = 1; // -> to temporarily enable debug output by default
  if(enableDebug) std::cout << "Processing TecModule::mainLoop()" << std::endl;

  tecIface = std::make_shared<TecChimeraInterface>(
      triggerCommands, responseStatus, commandMemoryRaw_wr, commandMemoryRaw_rd, true);
  tec = TecController(tecIface);

  // Make sure there is a proper FPGA communication timeout set if none provided by default
  if(fpgaTimeout < 100) {
    fpgaTimeout = 100;
    fpgaTimeout.write();
  }

  // Create TransferElementID to TecParameter maps
  createTransferIdGroupMaps();
  // createWriteBackAccessorsUnorderedSet();

  std::unordered_set<ChimeraTK::TransferElementID> writeBackAccessors;
  auto inserter = [&writeBackAccessors](auto& elem) { writeBackAccessors.insert(elem.second.getId()); };
  std::for_each(writeIntMap.begin(), writeIntMap.end(), inserter);
  std::for_each(writeFloatMap.begin(), writeFloatMap.end(), inserter);
  /*
  if(writeBackAccessorsLocal == writeBackAccessors)
    std::cout << "SETS ARE THE SAME !!!!!!!!!!!!!!!" << std::endl;
  else
    std::cout << "SETS ARE NOT THE SAME !!!!!!!!!!!!!!!" << std::endl;
  */
  // std::unordered_set<ChimeraTK::TransferElementID> writeBackAccessors;
  // for(auto& elem : writeIntMap) writeBackAccessors.insert(elem.second.getId());
  // for(auto& elem : writeFloatMap) writeBackAccessors.insert(elem.second.getId());

  errorlog.reset();
  // send out error log, for either the initial, empty, value or the updates provided by the functions below
  errorlog.write();

  // Read all parameters on module start
  std::cout << "Starting initial read-out of all parameters" << std::endl;
  for(auto& entry : readGroupMap) {
    readGroup(entry.second);
  }
  std::cout << "Initial read-out of all parameters is done" << std::endl;

  auto group = readAnyGroup();
  while(true) {
    auto id = group.readAny();

    if(id == errorlog.errorReset.getId()) {
      std::cout << "Error reset requested" << std::endl;
      errorlog.reset();
    }
    else if(id == fpgaTimeout.getId()) {
      if(fpgaTimeout < 100) {
        fpgaTimeout = 100;
        fpgaTimeout.write();
      }
    }
    else if(id == periodicRead.trigger.getId()) {
      readPeriodic();
    }
    else if(readGroupMap.find(id) != readGroupMap.end()) {
      readGroup(readGroupMap.at(id));
    }
    else if(writeGroupMap.find(id) != writeGroupMap.end()) {
      writeGroup(writeGroupMap.at(id));
    }
    else if(writeGroupToFlashMap.find(id) != writeGroupToFlashMap.end()) {
      writeGroupToFlash(writeGroupToFlashMap.at(id));
    }
    else if(setInputFromOutputGroupMap.find(id) != setInputFromOutputGroupMap.end()) {
      setInputFromOutputGroup(setInputFromOutputGroupMap.at(id));
    }
    else if(resetDevice.getId() == id) {
      doResetDevice();
    }
    else if(writeBackAccessors.count(id) == 0) {
      std::cout << "Unexpected input (id = " << id << ") has been written, don't know what to do!" << std::endl;
    }
    errorlog.write();
  }
}

/**********************************************************************************************************************/

void TecModule::readPeriodic() {
  if(enableDebug) std::cout << "readPeriodic()" << std::endl;

  std::vector<TecFrame> readRequests;
  for(auto& entry : enablePeriodicReadList) {
    if(entry.first != 0) {
      if(enableDebug) std::cout << "Querying value for " << entry.second << std::endl;
      readRequests.emplace_back(0, entry.second, TecFrame::TecFrameType::TecParameterValueRead, 1);
    }
  }

  if(enableHeartbeat && readRequests.size() == 0)
    readRequests.emplace_back(0, TecFrame::DeviceType, TecFrame::TecFrameType::TecParameterValueRead, 1);

  auto answers = doTecCommunication(readRequests);

  size_t index = 0;
  for(auto& frame : answers) {
    auto parameterOffset = readRequests[index++].getParameterOffset();
    try {
      auto& accessor = readIntMap.at(parameterOffset);
      checkAndUpdateReadResponses(frame.get(), accessor);
      continue;
    }
    catch(std::out_of_range&) {
      // Intentionally left blank, could be float
    }

    try {
      auto& accessor = readFloatMap.at(parameterOffset);
      checkAndUpdateReadResponses(frame.get(), accessor);
    }
    catch(std::out_of_range&) {
      // FIXME: Once done with mapping, throw exception?
      std::cout << "Unmapped TecParameter " << parameterOffset << ", skipping" << std::endl;
    }
  }
}

/**********************************************************************************************************************/

void TecModule::readGroup(std::list<TecFrame::TecParameter>& group) {
  if(enableDebug) std::cout << "readGroup()" << std::endl;
  std::vector<TecFrame> readRequests;
  for(auto& param : group) {
    if(enableDebug) std::cout << "Add read frame for param: " << param << std::endl;
    readRequests.emplace_back(0, param, TecFrame::TecFrameType::TecParameterValueRead, 1);
  }

  auto answers = doTecCommunication(readRequests);

  size_t index = 0;
  for(auto& frame : answers) {
    auto parameterOffset = readRequests[index++].getParameterOffset();

    if(readIntMap.find(parameterOffset) != readIntMap.end()) {
      auto& accessor = readIntMap.at(parameterOffset);
      checkAndUpdateReadResponses(frame.get(), accessor);
    }
    else {
      auto& accessor = readFloatMap.at(parameterOffset);
      checkAndUpdateReadResponses(frame.get(), accessor);
    }
  }
  if(enableDebug) std::cout << "done." << std::endl;
}

/**********************************************************************************************************************/

void TecModule::doResetDevice() {
  std::vector<TecFrame> resetRequests;
  if(enableDebug) std::cout << "Add reset frame" << std::endl;
  resetRequests.emplace_back(0, TecFrame::TecFrameType::TecReset);

  auto answers = doTecCommunication(resetRequests);
  auto* frame = answers[0].get();
  if(frame == nullptr) {
    errorlog.append("Hardware response Timeout limit for triggering device reset.");
    return;
  }

  if(enableDebug) {
    std::cout << "STATE = " << frame->getState() << std::endl;
    frame->print();
  }

  if(frame->getState() == TecFrame::TecFrameState::TecFrameParsed) {
    if(enableDebug) {
      std::cout << "Device reset successfully triggered" << std::endl;
    }
  }
  else {
    if(enableDebug) {
      std::cout << "Error: Frame not parsed, skipping" << std::endl;
    }
    errorlog.append("Device reset failed: " + getErrorString(*frame));
  }
}

/**********************************************************************************************************************/

template<typename T>
void TecModule::checkAndUpdateReadResponses(TecFrame* frame, ctk::ScalarOutput<T>& accessor) {
  if(frame == nullptr) {
    errorlog.append("Hardware response Timeout limit for read on " + accessor.getName() + " exceeded.");
    accessor.setDataValidity(ctk::DataValidity::faulty);
    accessor.write();
    return;
  }

  if(enableDebug) {
    std::cout << "STATE = " << frame->getState() << std::endl;
    frame->print();
  }

  if(frame->getState() == TecFrame::TecFrameState::TecFrameParsed) {
    if(accessor.getValueType() == typeid(int)) {
      accessor = frame->getIntParameterValue();
    }
    else {
      accessor = frame->getFloatParameterValue();
    }
    accessor.setDataValidity(ctk::DataValidity::ok);
    accessor.write();
    if(enableDebug) {
      std::cout << "Setting " + std::string(accessor.getValueType().name()) + " paramter "
                << frame->getParameterOffset() << " to " << accessor << std::endl;
    }
  }
  else {
    if(enableDebug) {
      std::cout << "Error: Frame not parsed, skipping" << std::endl;
    }
    errorlog.append("Read on " + accessor.getName() + " failed: " + getErrorString(*frame));
    accessor.setDataValidity(ctk::DataValidity::faulty);
    accessor.write();
  }
}

/**********************************************************************************************************************/

void TecModule::writeGroup(std::list<TecFrame::TecParameter>& group) {
  if(enableDebug) std::cout << "writeGroup()" << std::endl;

  // create request to disable write to flash
  std::vector<TecFrame> writeRequests;
  writeRequests.emplace_back(0, TecFrame::ParameterSystemFlashSaveOff, TecFrame::TecParameterValueSet, 1, (int)1);

  writeGroupInternal(group, writeRequests);
}

/**********************************************************************************************************************/

void TecModule::writeGroupToFlash(std::list<TecFrame::TecParameter>& group) {
  if(enableDebug) std::cout << "writeGroupToFlash()" << std::endl;

  // create request to enable write to flash
  std::vector<TecFrame> writeRequests;
  writeRequests.emplace_back(0, TecFrame::ParameterSystemFlashSaveOff, TecFrame::TecParameterValueSet, 1, (int)0);

  writeGroupInternal(group, writeRequests);
}

/**********************************************************************************************************************/
void TecModule::writeGroupInternal(std::list<TecFrame::TecParameter>& group, std::vector<TecFrame>& writeRequests) {
  for(auto& param : group) {
    try {
      auto& accessor = writeIntMap.at(param);
      writeRequests.emplace_back(0, param, TecFrame::TecFrameType::TecParameterValueSet, 1, accessor);
      continue;
    }
    catch(std::out_of_range&) {
      // Intentionally left blank
    }

    try {
      auto& accessor = writeFloatMap.at(param);
      writeRequests.emplace_back(0, param, TecFrame::TecFrameType::TecParameterValueSet, 1, accessor);
    }
    catch(std::out_of_range&) {
      std::cout << "Unmapped TecParameter " << param << ", skipping" << std::endl;
    }
  }

  auto answers = doTecCommunication(writeRequests);

  size_t index = 0;
  for(auto& f : answers) {
    auto parameterOffset = writeRequests[index++].getParameterOffset();
    try {
      auto& accessor = readIntMap.at(parameterOffset);
      checkAndUpdateWriteResponses(f.get(), accessor);
      continue;
    }
    catch(std::out_of_range&) {
      // Intentionally left blank
    }
    try {
      auto& accessor = readFloatMap.at(parameterOffset);
      checkAndUpdateWriteResponses(f.get(), accessor);
      continue;
    }
    catch(std::out_of_range&) {
      std::cout << "Unmapped TecParameter " << parameterOffset << ", skipping" << std::endl;
    }
  }
}
/**********************************************************************************************************************/

template<typename T>
void TecModule::checkAndUpdateWriteResponses(TecFrame* frame, ctk::ScalarOutput<T>& accessor) {
  if(frame == nullptr) {
    errorlog.append("Hardware communication timeout limit on write response for " + accessor.getName() + " exceeded.");
    return;
  }
  if(frame->getState() == TecFrame::TecFrameState::TecFrameParsed) {
    if(accessor.getValueType() == typeid(int)) {
      accessor = frame->getIntParameterValue();
    }
    else {
      accessor = frame->getFloatParameterValue();
    }
    accessor.write();
  }
  else {
    if(enableDebug) {
      std::cout << "Error: Frame not parsed, skipping" << std::endl;
    }
    errorlog.append("Write on " + accessor.getName() + " failed: " + getErrorString(*frame));
  }
}

/**********************************************************************************************************************/

void TecModule::setInputFromOutputGroup(std::list<TecFrame::TecParameter>& group) {
  for(auto& param : group) {
    if(readIntMap.find(param) != readIntMap.end()) {
      auto& readAccessor = readIntMap.at(param);
      assert(writeIntMap.find(param) != writeIntMap.end());
      auto& writeAccessor = writeIntMap.at(param);

      writeAccessor = static_cast<int>(readAccessor);
      writeAccessor.write();

      if(enableDebug) {
        std::cout << "Copying int value from " << readAccessor.getName() << " to " << writeAccessor.getName()
                  << std::endl;
      }
    }
    else {
      assert(readFloatMap.find(param) != readFloatMap.end());
      assert(writeFloatMap.find(param) != writeFloatMap.end());

      auto& readAccessor = readFloatMap.at(param);
      auto& writeAccessor = writeFloatMap.at(param);

      writeAccessor = static_cast<float>(readAccessor);
      writeAccessor.write();

      if(enableDebug) {
        std::cout << "Copying float value from " << readAccessor.getName() << " to " << writeAccessor.getName()
                  << std::endl;
      }
    }
  }
}

/**********************************************************************************************************************/

std::vector<std::shared_ptr<TecFrame>> TecModule::doTecCommunication(std::vector<TecFrame>& requests) {
  // No frames passed, nothing to do
  if(requests.size() == 0) return {};

  std::vector<std::shared_ptr<TecFrame>> answers;

  // add frames
  tec.clearAllFrames();
  auto requestIt = requests.begin();
  while(true) {
    bool commitTransaction = false;
    if(tec.addFrame(*requestIt)) {
      if(enableDebug) std::cout << "Frame added for parameter " << requestIt->getParameterOffset() << std::endl;
      if(++requestIt == requests.end()) commitTransaction = true;
    }
    else {
      // frame was not added because maximum number of frames was exceeded
      commitTransaction = true;
      if(enableDebug) std::cout << "Maximum number of frames reached, need to split transaction." << std::endl;
    }
    if(commitTransaction) {
      if(enableDebug) std::cout << "constructAllFrames:" << std::endl;
      tec.constructAllFrames();
      if(enableDebug) {
        for(uint8_t i = 0; i < tec.getFrameCount(); i++) {
          tec.getFrameAt(i)->print();
        }
        std::cout << "copyCommandsToHardware:" << std::endl;
      }
      tec.copyCommandsToHardware();
      if(enableDebug) std::cout << "triggerHardwareCommands:" << std::endl;
      tec.triggerHardwareCommands();

      if(enableDebug) std::cout << "readHardwareResponses:" << std::endl;
      uint32_t msPassed{0};
      bool timedOut{false};
      while(not tec.readHardwareResponses()) {
        if(msPassed == fpgaTimeout) {
          const std::string timeoutMessage = "Timeout limit for hardware response exceeded!";
          if(enableDebug) std::cout << "  " + timeoutMessage << std::endl;

          timedOut = true;
          break;
        }
        usleep(1000);
        ++msPassed;
      }

      // In case of timeout, we don't have a valid response
      if(not timedOut) {
        if(enableDebug) std::cout << "copyResponsesFromHardware:" << std::endl;
        tec.copyResponsesFromHardware();
        if(enableDebug) std::cout << "parseAllFrames:" << std::endl;
        tec.parseAllFrames();
        for(auto i = 0U; i < tec.getFrameCount(); i++) {
          answers.push_back(tec.getFrameAt(i));
        }
      }
      else { // timed out
        tec.resetCommunication();
        for(auto i = 0U; i < tec.getFrameCount(); i++) {
          answers.push_back(nullptr);
        }
      }

      // remove all frames for a fresh start of the next transaction
      tec.clearAllFrames();

      // bail out if at end of request list
      if(requestIt == requests.end()) break;
    }
  }
  assert(answers.size() == requests.size());
  return answers;
}

/**********************************************************************************************************************/

// FIXME: - think getErrorString should ideally be a TecFrame class method
//        - would be nice to mark TecFrame::getState as const.
//          could not use the const qualifier here because method is not
//          marked const
//        - Nice if TecFrame::getErrorCode could return TecErrorCode enum
//
std::string TecModule::getErrorString(TecFrame& f) {
  switch(f.getState()) {
    case TecFrame::TecFrameCRCError:
      return "Failed CRC check on frame.";
    case TecFrame::TecFrameTimedOut:
      return "Tec frame timed out";
    case TecFrame::TecFrameWithError:
      return TecErrorDescriptions.at(static_cast<TecErrorCode>(f.getErrorCode()));
    default:
      return "";
  }
}

/**********************************************************************************************************************/

void TecModule::ErrorLog::append(const std::string& e) {
  errorStatus = true;
  errorMessage = errorMessage + e + "\n";
  dirty = true;
}

/**********************************************************************************************************************/

void TecModule::ErrorLog::reset() {
  errorStatus = 0;
  errorMessage = "";
  dirty = true;
}

/**********************************************************************************************************************/

void TecModule::ErrorLog::write() {
  if(not dirty) {
    return;
  }
  errorStatus.write();
  errorMessage.write();
  dirty = false;
}

/**********************************************************************************************************************/

/**
 * Iterate over descriptions and maps group names to their respective trigger command accessors
 */
void TecModule::createCommandGroupMaps() {
  for(auto& d : descriptors) {
    for(const auto& group : {d.readGroup, d.writeGroup, d.writeToFlashGroup, d.setFromReadbackGroup}) {
      if(group.empty()) continue;
      groupCommandMap.emplace(
          group, ChimeraTK::ScalarPushInput<int>{this, group + "_wr", "", detail::generateGroupDescription(group)});
    }
  }
}

/**********************************************************************************************************************/
void TecModule::createReadWriteMaps() {
  auto insertIntoMaps = [this](ParameterDescriptor& d) {
    if(d.valueType == ValueType::Int) {
      readIntMap.emplace(d.value, ctk::ScalarOutput<int>{this, d.pvName + "_rd", d.unit, d.description});

      if(d.direction == Direction::ReadWrite) {
        writeIntMap.emplace(d.value, ctk::ScalarPushInputWB<int>{this, d.pvName + "_wr", d.unit, d.description});
      }
    }
    else if(d.valueType == ValueType::Float) {
      readFloatMap.emplace(d.value, ctk::ScalarOutput<float>{this, d.pvName + "_rd", d.unit, d.description});

      if(d.direction == Direction::ReadWrite) {
        writeFloatMap.emplace(d.value, ctk::ScalarPushInputWB<float>{this, d.pvName + "_wr", d.unit, d.description});
      }
    }
  };
  std::for_each(descriptors.begin(), descriptors.end(), insertIntoMaps);
}

/**********************************************************************************************************************/
void TecModule::createEnablePeriodicList() {
  // TODO: Should this be combined with the implementation of #6567?
  for(auto& desc : descriptors) {
    if(desc.direction != Direction::Read) continue;

    enablePeriodicReadList.emplace_back(ChimeraTK::ScalarPollInput<int>(this, desc.pvName + "_enaPeriodicRead", "",
                                            "Enable periodic read-out of " + desc.pvName),
        desc.value);
  }
}

/**
 * Iterates over all ParameterDescriptors and maps the TransferElementID of
 * the related input in groupCommandMap to the TecParameter value
 */
void TecModule::createTransferIdGroupMaps() {
  for(auto& descriptor : descriptors) {
    try {
      if(!descriptor.readGroup.empty()) {
        readGroupMap[groupCommandMap.at(descriptor.readGroup).getId()].emplace_back(descriptor.value);
      }
      if(!descriptor.writeGroup.empty()) {
        writeGroupMap[groupCommandMap.at(descriptor.writeGroup).getId()].emplace_back(descriptor.value);
      }
      if(!descriptor.writeToFlashGroup.empty()) {
        writeGroupToFlashMap[groupCommandMap.at(descriptor.writeToFlashGroup).getId()].emplace_back(descriptor.value);
      }
      if(!descriptor.setFromReadbackGroup.empty()) {
        setInputFromOutputGroupMap[groupCommandMap.at(descriptor.setFromReadbackGroup).getId()].emplace_back(
            descriptor.value);
      }
    }
    catch(std::out_of_range&) {
      std::cout << "Missing accessor in groupCommandMap! This warning should later become a hard error!" << std::endl;
    }
  }
}

void TecModule::createWriteBackAccessorsUnorderedSet() {
  // for(auto& elem : writeIntMap) writeBackAccessors.insert(elem.second.getId());
  // for(auto& elem : writeFloatMap) writeBackAccessors.insert(elem.second.getId());
}
TecModule::TecModule(TecModule&& other) noexcept
: ctk::ApplicationModule(std::move(other)), tecIface(std::move(other.tecIface)), tec(other.tec) {}
