// Define a name for the test module.
#define BOOST_TEST_MODULE testRead
// Only after defining the name include the unit test header.
#include <boost/test/included/unit_test.hpp>

#include <ChimeraTK/ApplicationCore/TestFacility.h>
#include "Server.h"
#include "TecDummy.h"

using namespace boost::unit_test_framework;
using namespace ChimeraTK;

BOOST_AUTO_TEST_SUITE(tecTestSuite)

constexpr char TECDummyCDD[] = "(TecDummy:?map=dummydevice.map)";

BOOST_AUTO_TEST_CASE(testReadOnDemandInteger) {
  std::cout << "testReadOnDemandInteger" << std::endl;

  Server tec_app{"tec_dummy"};
  ChimeraTK::TestFacility testFacility;

  testFacility.runApplication();

  boost::shared_ptr<TecDummy> TECDummy =
      boost::dynamic_pointer_cast<TecDummy>(ChimeraTK::BackendFactory::getInstance().createBackend(TECDummyCDD));

  testFacility.writeScalar("TecModule0/readGroup1_wr", 1);
  testFacility.stepApplication();
  auto serialNumber = testFacility.readScalar<int>("TecModule0/mon_serialNumber_rd");
  BOOST_CHECK_EQUAL(serialNumber, 0);

  auto setSerialNumberValue = 9;
  for(auto i = 0; i < 6; i++) {
    setSerialNumberValue = setSerialNumberValue * 13;
    TECDummy->set(TecFrame::TecParameter::MonSerialNumber, setSerialNumberValue);
    testFacility.writeScalar("TecModule0/readGroup1_wr", 2);
    testFacility.stepApplication();
    auto serialNumberRead = testFacility.readScalar<int>("TecModule0/mon_serialNumber_rd");
    // std::cout<<"serialNumberRead:"<<serialNumberRead<<std::endl;
    BOOST_CHECK_EQUAL(serialNumberRead, setSerialNumberValue);
  }
}

BOOST_AUTO_TEST_CASE(testReadOnDemandFloat) {
  std::cout << "testReadOnDemandFloat" << std::endl;

  Server tec_app{"tec_dummy"};
  ChimeraTK::TestFacility testFacility;

  testFacility.runApplication();

  boost::shared_ptr<TecDummy> TECDummy =
      boost::dynamic_pointer_cast<TecDummy>(ChimeraTK::BackendFactory::getInstance().createBackend(TECDummyCDD));

  testFacility.writeScalar("TecModule0/readCh1Group1_wr", 1);
  testFacility.stepApplication();
  auto temperature = testFacility.readScalar<float_t>("TecModule0/mon_ch1_sinkTemp_rd");
  BOOST_CHECK_EQUAL(temperature, 0);

  auto setTemperatureValue = 0.009f;
  for(auto i = 0; i < 6; i++) {
    setTemperatureValue = setTemperatureValue * 11.0f;
    TECDummy->set(TecFrame::TecParameter::MonSinkTemperature, setTemperatureValue);
    testFacility.writeScalar("TecModule0/readCh1Group1_wr", 2);
    testFacility.stepApplication();
    auto temperatureRead = testFacility.readScalar<float_t>("TecModule0/mon_ch1_sinkTemp_rd");
    // std::cout<<"temperatureRead:"<<temperatureRead<<std::endl;
    BOOST_CHECK_CLOSE(temperatureRead, setTemperatureValue, 0.0001);
  }
}

BOOST_AUTO_TEST_CASE(testReadOnWritable) {
  std::cout << "testReadOnWritable" << std::endl;

  Server tec_app{"tec_dummy"};
  ChimeraTK::TestFacility testFacility;

  testFacility.runApplication();
  boost::shared_ptr<TecDummy> TECDummy =
      boost::dynamic_pointer_cast<TecDummy>(ChimeraTK::BackendFactory::getInstance().createBackend(TECDummyCDD));

  testFacility.writeScalar("TecModule0/readCh1WriteGroup7_wr", 1);
  testFacility.stepApplication();
  auto fanControl = testFacility.readScalar<int>("TecModule0/ctrl_ch1_fanControlEnable_rd");
  BOOST_CHECK_EQUAL(fanControl, 0);

  auto setFanControlValue = 7;
  for(auto i = 0; i < 6; i++) {
    setFanControlValue = setFanControlValue * 13;
    TECDummy->set(TecFrame::TecParameter::ExpFanControlEnable, setFanControlValue);
    testFacility.writeScalar("TecModule0/readCh1WriteGroup7_wr", 2);
    testFacility.stepApplication();
    auto fanControlRead = testFacility.readScalar<int>("TecModule0/ctrl_ch1_fanControlEnable_rd");
    // std::cout<<"fanControlRead:"<<fanControlRead<<std::endl;
    BOOST_CHECK_EQUAL(fanControlRead, setFanControlValue);
  }
}

BOOST_AUTO_TEST_CASE(testReadBackFromWrite) {
  std::cout << "testReadBackFromWrite" << std::endl;

  Server tec_app{"tec_dummy"};
  ChimeraTK::TestFacility testFacility;

  testFacility.runApplication();
  boost::shared_ptr<TecDummy> TECDummy =
      boost::dynamic_pointer_cast<TecDummy>(ChimeraTK::BackendFactory::getInstance().createBackend(TECDummyCDD));

  auto setFanControlValue = 3;
  for(auto i = 0; i < 6; i++) {
    setFanControlValue = setFanControlValue * 13;
    TECDummy->set(TecFrame::TecParameter::ExpFanControlEnable, setFanControlValue);
    testFacility.writeScalar("TecModule0/readCh1WriteGroup7_wr", 2);
    testFacility.stepApplication();
    auto fanControlRead = testFacility.readScalar<int>("TecModule0/ctrl_ch1_fanControlEnable_rd");
    // std::cout<<"fanControlRead:"<<fanControlRead<<"::setFanControlValue:"<<setFanControlValue<<std::endl;
    testFacility.writeScalar("TecModule0/setRbCh1Group7_wr", 2);
    testFacility.stepApplication();
    auto fanControlReadBack = testFacility.readScalar<int>("TecModule0/ctrl_ch1_fanControlEnable_wr");
    // std::cout<<"fanControlRead:"<<fanControlRead<<"::setFanControlValue:"<<setFanControlValue<<std::endl;
    // std::cout<<"fanControlRead:"<<fanControlRead<<"::fanControlReadBack:"<<fanControlReadBack<<std::endl;
    BOOST_CHECK_EQUAL(fanControlRead, fanControlReadBack);
  }
  setFanControlValue = -3;
  for(auto i = 0; i < 6; i++) {
    setFanControlValue = setFanControlValue * 17;
    TECDummy->set(TecFrame::TecParameter::ExpFanControlEnable, setFanControlValue);
    testFacility.writeScalar("TecModule0/readCh1WriteGroup7_wr", 2);
    testFacility.stepApplication();
    auto fanControlRead = testFacility.readScalar<int>("TecModule0/ctrl_ch1_fanControlEnable_rd");
    // std::cout<<"fanControlRead:"<<fanControlRead<<"::setFanControlValue:"<<setFanControlValue<<std::endl;
    testFacility.writeScalar("TecModule0/setRbCh1Group7_wr", 2);
    testFacility.stepApplication();
    auto fanControlReadBack = testFacility.readScalar<int>("TecModule0/ctrl_ch1_fanControlEnable_wr");
    // std::cout<<"fanControlRead:"<<fanControlRead<<"::setFanControlValue:"<<setFanControlValue<<std::endl;
    // std::cout<<"fanControlRead:"<<fanControlRead<<"::fanControlReadBack:"<<fanControlReadBack<<std::endl;
    BOOST_CHECK_EQUAL(fanControlRead, fanControlReadBack);
  }
}

BOOST_AUTO_TEST_SUITE_END()
