// Define a name for the test module.
#define BOOST_TEST_MODULE testWrite
// Only after defining the name include the unit test header.
#include <boost/test/included/unit_test.hpp>
#include <ChimeraTK/ApplicationCore/TestFacility.h>
#include <ChimeraTK/BackendFactory.h>
#include "Server.h"
#include "TecDummy.h"

using namespace boost::unit_test_framework;

struct Fixture_t {
  Server theServer{"tec_dummy"};
  ChimeraTK::TestFacility testFacility;
  boost::shared_ptr<TecDummy> dummy{
      boost::dynamic_pointer_cast<TecDummy>(ChimeraTK::BackendFactory::getInstance().createBackend("ActualDevice"))};
};

/// TestSuite for the server, adapt name
BOOST_AUTO_TEST_SUITE(TecServerTestSuite)

/**********************************************************************************************************************/

/// Test write of an integer
BOOST_FIXTURE_TEST_CASE(testInteger, Fixture_t) {
  std::cout << "testInteger" << std::endl;

  auto fanControlEnable = testFacility.getScalar<int>("TecModule0/ctrl_ch1_fanControlEnable_wr");
  auto writeCh1Group7 = testFacility.getScalar<int>("TecModule0/writeCh1Group7_wr");

  testFacility.runApplication();

  fanControlEnable = 1;
  fanControlEnable.write();
  writeCh1Group7.write();
  BOOST_CHECK_EQUAL(dummy->getInt(TecFrame::TecParameter::ExpFanControlEnable), 0);
  testFacility.stepApplication();
  BOOST_CHECK_EQUAL(dummy->getInt(TecFrame::TecParameter::ExpFanControlEnable), 1);

  fanControlEnable = 42;
  fanControlEnable.write();
  writeCh1Group7.write();
  BOOST_CHECK_EQUAL(dummy->getInt(TecFrame::TecParameter::ExpFanControlEnable), 1);
  testFacility.stepApplication();
  BOOST_CHECK_EQUAL(dummy->getInt(TecFrame::TecParameter::ExpFanControlEnable), 42);

  fanControlEnable = -120;
  fanControlEnable.write();
  writeCh1Group7.write();
  BOOST_CHECK_EQUAL(dummy->getInt(TecFrame::TecParameter::ExpFanControlEnable), 42);
  testFacility.stepApplication();
  BOOST_CHECK_EQUAL(dummy->getInt(TecFrame::TecParameter::ExpFanControlEnable), -120);
}

/**********************************************************************************************************************/

/// Test write of a float
BOOST_FIXTURE_TEST_CASE(testFloat, Fixture_t) {
  std::cout << "testFloat" << std::endl;

  auto targetObjectTemp = testFacility.getScalar<float>("TecModule0/ctrl_ch1_targetObjectTemp_wr");
  auto writeCh1Group1 = testFacility.getScalar<int>("TecModule0/writeCh1Group1_wr");

  testFacility.runApplication();

  targetObjectTemp = 3.1415;
  targetObjectTemp.write();
  writeCh1Group1.write();
  BOOST_CHECK_CLOSE(dummy->getFloat(TecFrame::TecParameter::TemTargetObjectTemp), 0.0, 0.0001);
  testFacility.stepApplication();
  BOOST_CHECK_CLOSE(dummy->getFloat(TecFrame::TecParameter::TemTargetObjectTemp), 3.1415, 0.0001);

  targetObjectTemp = 3.5e27;
  targetObjectTemp.write();
  writeCh1Group1.write();
  BOOST_CHECK_CLOSE(dummy->getFloat(TecFrame::TecParameter::TemTargetObjectTemp), 3.1415, 0.0001);
  testFacility.stepApplication();
  BOOST_CHECK_CLOSE(dummy->getFloat(TecFrame::TecParameter::TemTargetObjectTemp), 3.5e27, 0.0001);

  targetObjectTemp = -120;
  targetObjectTemp.write();
  writeCh1Group1.write();
  BOOST_CHECK_CLOSE(dummy->getFloat(TecFrame::TecParameter::TemTargetObjectTemp), 3.5e27, 0.0001);
  testFacility.stepApplication();
  BOOST_CHECK_CLOSE(dummy->getFloat(TecFrame::TecParameter::TemTargetObjectTemp), -120, 0.0001);
}

/**********************************************************************************************************************/

BOOST_FIXTURE_TEST_CASE(testFlashWrite, Fixture_t) {
  std::cout << "testWriteFlash" << std::endl;

  auto fanControlEnable = testFacility.getScalar<int>("TecModule0/ctrl_ch1_fanControlEnable_wr");
  auto fanTargetTemp = testFacility.getScalar<float>("TecModule0/ctrl_ch1_fanTargetTemp_wr");
  auto writeCh1Group7 = testFacility.getScalar<int>("TecModule0/writeCh1Group7_wr");
  auto writeToFlashCh1Group7 = testFacility.getScalar<int>("TecModule0/writeToFlashCh1Group7_wr");

  testFacility.runApplication();

  fanControlEnable = 1;
  fanTargetTemp = 42.0;
  fanControlEnable.write();
  fanTargetTemp.write();
  writeCh1Group7.write();
  testFacility.stepApplication();

  BOOST_CHECK_EQUAL(dummy->getInt(TecFrame::TecFrame::ExpFanControlEnable), 1);
  BOOST_CHECK_CLOSE(dummy->getFloat(TecFrame::TecParameter::ExpFanTargetTemp), 42.0, 1e5);
  BOOST_CHECK(not dummy->getLastWriteToFlash(TecFrame::TecFrame::ExpFanControlEnable));
  BOOST_CHECK(not dummy->getLastWriteToFlash(TecFrame::TecFrame::ExpFanTargetTemp));

  fanControlEnable = 0;
  fanTargetTemp = 21.0;
  fanControlEnable.write();
  fanTargetTemp.write();
  writeToFlashCh1Group7.write();
  testFacility.stepApplication();
  BOOST_CHECK_EQUAL(dummy->getInt(TecFrame::TecFrame::ExpFanControlEnable), 0);
  BOOST_CHECK_CLOSE(dummy->getFloat(TecFrame::TecParameter::ExpFanTargetTemp), 21.0, 1e5);
  BOOST_CHECK(dummy->getLastWriteToFlash(TecFrame::TecFrame::ExpFanControlEnable));
  BOOST_CHECK(dummy->getLastWriteToFlash(TecFrame::TecFrame::ExpFanTargetTemp));

  writeCh1Group7.write();
  testFacility.stepApplication();
  BOOST_CHECK(not dummy->getLastWriteToFlash(TecFrame::TecFrame::ExpFanControlEnable));
  BOOST_CHECK(not dummy->getLastWriteToFlash(TecFrame::TecFrame::ExpFanTargetTemp));
}

/**********************************************************************************************************************/

BOOST_AUTO_TEST_SUITE_END()
